@@ -8,7 +8,6 @@
 
 use std::{
     collections::HashSet,
-    ffi::{CStr, CString},
     fs::{create_dir_all, File, OpenOptions},
     io::{BufRead, BufReader, Write},
     os::unix::fs::PermissionsExt,
@@ -17,7 +16,6 @@ use std::{
 
 use chrono::{NaiveDateTime, TimeZone};
 use log::{debug, error, warn};
-use scopeguard::defer;
 
 use crate::{Error, TrashContext, TrashItem};
 
@@ -75,11 +73,11 @@ pub fn list() -> Result<Vec<TrashItem>, Error> {
         }
     }
 
-    // Get all mountpoints and attemt to find a trash folder in each adding them to the SET of
+    // Get all mount-points and attempt to find a trash folder in each adding them to the SET of
     // trash folders when found one.
     let uid = unsafe { libc::getuid() };
     let mount_points = get_mount_points()?;
-    for mount in mount_points.into_iter() {
+    for mount in &mount_points {
         execute_on_mounted_trash_folders(uid, &mount.mnt_dir, false, false, |trash_path| {
             trash_folders.insert(trash_path);
             Ok(())
@@ -96,7 +94,7 @@ pub fn list() -> Result<Vec<TrashItem>, Error> {
     let mut result = Vec::new();
     for folder in trash_folders.iter() {
         // Read the info files for every file
-        let trash_folder_parent = folder.parent().unwrap();
+        let top_dir = get_topdir_for_path(folder, &mount_points);
         let info_folder = folder.join("info");
         if !info_folder.is_dir() {
             warn!(
@@ -119,39 +117,36 @@ pub fn list() -> Result<Vec<TrashItem>, Error> {
             }
         };
         'trash_item: for entry in read_dir {
-            let info_entry;
-            match entry {
-                Ok(entry) => info_entry = entry,
+            let info_entry = match entry {
+                Ok(entry) => entry,
                 Err(e) => {
                     // Another thread or process may have removed that entry by now
                     debug!("Tried resolving the trash info `DirEntry` but it failed with: '{}'", e);
                     continue;
                 }
-            }
-            // Entrt should really be an info file but better safe than sorry
-            let file_type;
-            match info_entry.file_type() {
-                Ok(f_type) => file_type = f_type,
+            };
+            // Entrty should really be an info file but better safe than sorry
+            let file_type = match info_entry.file_type() {
+                Ok(f_type) => f_type,
                 Err(e) => {
                     // Another thread or process may have removed that entry by now
                     debug!("Tried getting the file type of the trash info `DirEntry` but failed with: {}", e);
                     continue;
                 }
-            }
+            };
             let info_path = info_entry.path();
             if !file_type.is_file() {
                 warn!("Found an item that's not a file, among the trash info files. This is unexpected. The path to the item is: '{:?}'", info_path);
                 continue;
             }
-            let info_file;
-            match File::open(&info_path) {
-                Ok(file) => info_file = file,
+            let info_file = match File::open(&info_path) {
+                Ok(file) => file,
                 Err(e) => {
                     // Another thread or process may have removed that entry by now
                     debug!("Tried opening the trash info '{:?}' but failed with: {}", info_path, e);
                     continue;
                 }
-            }
+            };
             let id = info_path.clone().into();
             let mut name = None;
             let mut original_parent: Option<PathBuf> = None;
@@ -175,7 +170,7 @@ pub fn list() -> Result<Vec<TrashItem>, Error> {
                 if key == "Path" {
                     let mut value_path = Path::new(value).to_owned();
                     if value_path.is_relative() {
-                        value_path = trash_folder_parent.join(value_path);
+                        value_path = top_dir.join(value_path);
                     }
                     let full_path_utf8 = PathBuf::from(parse_uri_path(&value_path));
                     name = Some(full_path_utf8.file_name().unwrap().to_str().unwrap().to_owned());
@@ -230,10 +225,7 @@ where
         // A bunch of unwraps here. This is fine because if any of these fail that means
         // that either there's a bug in this code or the target system didn't follow
         // the specification.
-        let trash_folder = Path::new(info_file).parent().unwrap().parent().unwrap();
-        let name_in_trash = Path::new(info_file).file_stem().unwrap();
-
-        let file = trash_folder.join("files").join(&name_in_trash);
+        let file = restorable_file_in_trash_from_info_file(info_file);
         assert!(file.exists());
         if file.is_dir() {
             std::fs::remove_dir_all(&file).map_err(|e| fsys_err_to_unknown(&file, e))?;
@@ -247,6 +239,13 @@ where
     Ok(())
 }
 
+fn restorable_file_in_trash_from_info_file(info_file: impl AsRef<std::ffi::OsStr>) -> PathBuf {
+    let info_file = info_file.as_ref();
+    let trash_folder = Path::new(info_file).parent().unwrap().parent().unwrap();
+    let name_in_trash = Path::new(info_file).file_stem().unwrap();
+    trash_folder.join("files").join(&name_in_trash)
+}
+
 pub fn restore_all<I>(items: I) -> Result<(), Error>
 where
     I: IntoIterator<Item = TrashItem>,
@@ -263,10 +262,7 @@ where
         // A bunch of unwraps here. This is fine because if any of these fail that means
         // that either there's a bug in this code or the target system didn't follow
         // the specification.
-        let trash_folder = Path::new(info_file).parent().unwrap().parent().unwrap();
-        let name_in_trash = Path::new(info_file).file_stem().unwrap();
-
-        let file = trash_folder.join("files").join(&name_in_trash);
+        let file = restorable_file_in_trash_from_info_file(info_file);
         assert!(file.exists());
         // TODO add option to forcefully replace any target at the restore location
         // if it already exists.
@@ -503,7 +499,17 @@ fn move_items_no_replace(
     try_creating_placeholders(src, dst)?;
 
     // All placeholders are in place. LET'S OVERWRITE
-    execute_src_to_dst_operation(src, dst, &|_| Ok(()), &|src, dst| std::fs::rename(src, dst))?;
+    execute_src_to_dst_operation(src, dst, &|_| Ok(()), &|src, dst| {
+        if let Some(parent) = dst.parent() {
+            if let Err(err) = std::fs::create_dir_all(parent) {
+                warn!(
+                    "Failed to create destination directory. It probably already exists. {:?}",
+                    err
+                );
+            }
+        }
+        std::fs::rename(src, dst)
+    })?;
 
     // Once everything is moved, lets recursively remove the directory
     if src.is_dir() {
@@ -627,7 +633,23 @@ struct MountPoint {
     _mnt_fsname: String,
 }
 
+#[cfg(target_os = "linux")]
 fn get_mount_points() -> Result<Vec<MountPoint>, Error> {
+    use once_cell::sync::Lazy;
+    use scopeguard::defer;
+    use std::ffi::{CStr, CString};
+    use std::sync::Mutex;
+
+    // The getmntinfo() function writes the array of structures to an internal
+    // static object and returns a pointer to that object.  Subsequent calls to
+    // getmntent() will modify the same object. This means that the function is
+    // not threadsafe. To help prevent multiple threads using it concurrently
+    // via get_mount_points a Mutex is used.
+    // We understand that threads can still call `libc::getmntent(…)` directly
+    // to bypass the lock and trigger UB.
+    static LOCK: Lazy<Mutex<()>> = Lazy::new(|| Mutex::new(()));
+    let _lock = LOCK.lock().unwrap();
+
     //let file;
     let read_arg = CString::new("r").unwrap();
     let mounts_path = CString::new("/proc/mounts").unwrap();
@@ -672,8 +694,67 @@ fn get_mount_points() -> Result<Vec<MountPoint>, Error> {
     Ok(result)
 }
 
+#[cfg(any(target_os = "freebsd", target_os = "openbsd", target_os = "netbsd"))]
+fn get_mount_points() -> Result<Vec<MountPoint>, Error> {
+    use once_cell::sync::Lazy;
+    use std::sync::Mutex;
+
+    // The getmntinfo() function writes the array of structures to an internal
+    // static object and returns a pointer to that object.  Subsequent calls to
+    // getmntinfo() will modify the same object. This means that the function is
+    // not threadsafe. To help prevent multiple threads using it concurrently
+    // via get_mount_points a Mutex is used.
+    // We understand that threads can still call `libc::getmntinfo(…)` directly
+    // to bypass the lock and trigger UB.
+    static LOCK: Lazy<Mutex<()>> = Lazy::new(|| Mutex::new(()));
+    let _lock = LOCK.lock().unwrap();
+
+    fn c_buf_to_str(buf: &[libc::c_char]) -> Option<&str> {
+        let buf: &[u8] = unsafe { std::slice::from_raw_parts(buf.as_ptr() as _, buf.len()) };
+        if let Some(pos) = buf.iter().position(|x| *x == 0) {
+            // Shrink buffer to omit the null bytes
+            std::str::from_utf8(&buf[..pos]).ok()
+        } else {
+            std::str::from_utf8(buf).ok()
+        }
+    }
+    let mut fs_infos: *mut libc::statfs = std::ptr::null_mut();
+    let count = unsafe { libc::getmntinfo(&mut fs_infos, libc::MNT_WAIT) };
+    if count < 1 {
+        return Ok(Vec::new());
+    }
+    let fs_infos: &[libc::statfs] =
+        unsafe { std::slice::from_raw_parts(fs_infos as _, count as _) };
+
+    let mut result = Vec::new();
+    for fs_info in fs_infos {
+        if fs_info.f_mntfromname[0] == 0 || fs_info.f_mntonname[0] == 0 {
+            // If we have missing information, no need to look any further...
+            continue;
+        }
+        let fs_type = c_buf_to_str(&fs_info.f_fstypename).unwrap_or_default();
+        let mount_to = match c_buf_to_str(&fs_info.f_mntonname) {
+            Some(m) => m,
+            None => {
+                debug!("Cannot get disk mount point, ignoring it.");
+                continue;
+            }
+        };
+        let mount_from = c_buf_to_str(&fs_info.f_mntfromname).unwrap_or_default();
+
+        let mount_point = MountPoint {
+            mnt_dir: mount_to.into(),
+            _mnt_fsname: mount_from.into(),
+            _mnt_type: fs_type.into(),
+        };
+        result.push(mount_point);
+    }
+    Ok(result)
+}
+
 #[cfg(test)]
 mod tests {
+    use serial_test::serial;
     use std::{
         collections::{hash_map::Entry, HashMap},
         env,
@@ -694,6 +775,7 @@ mod tests {
     };
 
     #[test]
+    #[serial]
     fn test_list() {
         crate::tests::init_logging();
 
